@@ -159,7 +159,7 @@ fn escape_cirru_str(s: &str) -> String {
   result
 }
 
-fn quote_to_js(xs: &Calcit, var_prefix: &str, keywords: &RefCell<Vec<EdnKwd>>) -> Result<String, String> {
+fn quote_to_js(xs: &Calcit, var_prefix: &str, keywords: &RefCell<HashSet<EdnKwd>>) -> Result<String, String> {
   match xs {
     Calcit::Symbol { sym, .. } => Ok(format!("new {}CalcitSymbol({})", var_prefix, escape_cirru_str(sym))),
     Calcit::Str(s) => Ok(escape_cirru_str(s)),
@@ -180,7 +180,7 @@ fn quote_to_js(xs: &Calcit, var_prefix: &str, keywords: &RefCell<Vec<EdnKwd>>) -
     }
     Calcit::Keyword(s) => {
       let mut kwds = keywords.borrow_mut();
-      kwds.push(s.to_owned());
+      kwds.insert(s.to_owned());
       Ok(format!("_kwd[{}]", escape_cirru_str(&s.to_string())))
     }
     _ => unreachable!("Unexpected data in quote for js: {}", xs),
@@ -204,7 +204,7 @@ fn to_js_code(
   ns: &str,
   local_defs: &HashSet<Arc<str>>,
   file_imports: &RefCell<ImportsDict>,
-  keywords: &RefCell<Vec<EdnKwd>>,
+  keywords: &RefCell<HashSet<EdnKwd>>,
   return_label: Option<&str>,
 ) -> Result<String, String> {
   if let Calcit::List(ys) = xs {
@@ -218,7 +218,13 @@ fn to_js_code(
         resolved,
       } => {
         let resolved_info = resolved.to_owned().map(|v| (*v).to_owned());
-        gen_symbol_code(sym, &**def_ns, &**at_def, resolved_info, ns, xs, local_defs, file_imports)
+        let passed_defs = PassedDefs {
+          ns,
+          local_defs,
+          file_imports,
+        };
+
+        gen_symbol_code(sym, &**def_ns, &**at_def, resolved_info, xs, &passed_defs)
       }
       Calcit::Proc(s) => {
         let proc_prefix = get_proc_prefix(ns);
@@ -251,7 +257,7 @@ fn to_js_code(
       Calcit::Nil => Ok(String::from("null")),
       Calcit::Keyword(s) => {
         let mut kwds = keywords.borrow_mut();
-        kwds.push(s.to_owned());
+        kwds.insert(s.to_owned());
         Ok(format!("_kwd[{}]", wrap_js_str(&s.to_string())))
       }
       Calcit::List(_) => unreachable!("[Error] list handled in another branch"),
@@ -271,7 +277,7 @@ fn gen_call_code(
   local_defs: &HashSet<Arc<str>>,
   xs: &Calcit,
   file_imports: &RefCell<ImportsDict>,
-  keywords: &RefCell<Vec<EdnKwd>>,
+  keywords: &RefCell<HashSet<EdnKwd>>,
   return_label: Option<&str>,
 ) -> Result<String, String> {
   let return_code = return_label.unwrap_or("");
@@ -341,7 +347,12 @@ fn gen_call_code(
           (Some(Calcit::Symbol { sym, .. }), Some(Calcit::List(ys))) => {
             let func_body = body.skip(2)?;
             gen_stack::push_call_stack(ns, sym, StackKind::Codegen, xs.to_owned(), &TernaryTreeList::Empty);
-            let ret = gen_js_func(sym, &get_raw_args(ys)?, &func_body, ns, false, local_defs, file_imports, keywords);
+            let passed_defs = PassedDefs {
+              ns,
+              local_defs,
+              file_imports,
+            };
+            let ret = gen_js_func(sym, &get_raw_args(ys)?, &func_body, &passed_defs, false, keywords);
             gen_stack::pop_call_stack();
             match ret {
               Ok(code) => Ok(format!("{}{}", return_code, code)),
@@ -564,18 +575,23 @@ fn gen_call_code(
   }
 }
 
+/// a group of arguments related to scopes
+struct PassedDefs<'a> {
+  ns: &'a str,
+  local_defs: &'a HashSet<Arc<str>>,
+  file_imports: &'a RefCell<ImportsDict>,
+}
+
 fn gen_symbol_code(
   s: &str,
   def_ns: &str,
   at_def: &str,
   resolved: Option<primes::SymbolResolved>,
-  ns: &str,
   xs: &Calcit,
-  local_defs: &HashSet<Arc<str>>,
-  file_imports: &RefCell<ImportsDict>,
+  passed_defs: &PassedDefs,
 ) -> Result<String, String> {
   // println!("gen symbol: {} {} {} {:?}", s, def_ns, ns, resolved);
-  let var_prefix = if ns == primes::CORE_NS { "" } else { "$calcit." };
+  let var_prefix = if passed_defs.ns == primes::CORE_NS { "" } else { "$calcit." };
   if has_ns_part(s) {
     if s.starts_with("js/") {
       Ok(escape_ns_var(s, "js"))
@@ -590,10 +606,10 @@ fn gen_symbol_code(
           rule: _import_rule, /* None */
         }) => {
           if is_cirru_string(&*r_ns) {
-            track_ns_import(ns_part, ImportedTarget::AsNs(r_ns), file_imports)?;
+            track_ns_import(ns_part, ImportedTarget::AsNs(r_ns), passed_defs.file_imports)?;
             Ok(escape_ns_var(s, ns_part))
           } else {
-            track_ns_import(&*r_ns, ImportedTarget::AsNs(r_ns.to_owned()), file_imports)?;
+            track_ns_import(&*r_ns, ImportedTarget::AsNs(r_ns.to_owned()), passed_defs.file_imports)?;
             Ok(escape_ns_var(s, &*r_ns))
           }
         }
@@ -604,9 +620,9 @@ fn gen_symbol_code(
     }
   } else if is_js_syntax_procs(s) || is_proc_name(s) || CalcitSyntax::is_core_syntax(s) {
     // return Ok(format!("{}{}", var_prefix, escape_var(s)));
-    let proc_prefix = get_proc_prefix(ns);
+    let proc_prefix = get_proc_prefix(passed_defs.ns);
     return Ok(format!("{}{}", proc_prefix, escape_var(s)));
-  } else if matches!(resolved, Some(ResolvedLocal)) || local_defs.contains(s) {
+  } else if matches!(resolved, Some(ResolvedLocal)) || passed_defs.local_defs.contains(s) {
     Ok(escape_var(s))
   } else if let Some(ResolvedDef {
     ns: r_ns,
@@ -620,9 +636,9 @@ fn gen_symbol_code(
     }
     if let Some(ImportRule::NsDefault(_s)) = import_rule.map(|x| (&*x).to_owned()) {
       // imports that using :default are special
-      track_ns_import(s, ImportedTarget::DefaultNs(r_ns), file_imports)?;
+      track_ns_import(s, ImportedTarget::DefaultNs(r_ns), passed_defs.file_imports)?;
     } else {
-      track_ns_import(s, ImportedTarget::ReferNs(r_ns), file_imports)?;
+      track_ns_import(s, ImportedTarget::ReferNs(r_ns), passed_defs.file_imports)?;
     }
     Ok(escape_var(s))
   } else if def_ns == primes::CORE_NS {
@@ -631,18 +647,18 @@ fn gen_symbol_code(
     Ok(format!("{}{}", var_prefix, escape_var(s)))
   } else if def_ns.is_empty() {
     Err(format!("Unexpected ns at symbol, {:?}", xs))
-  } else if def_ns != ns {
-    track_ns_import(s, ImportedTarget::ReferNs(def_ns.into()), file_imports)?;
+  } else if def_ns != passed_defs.ns {
+    track_ns_import(s, ImportedTarget::ReferNs(def_ns.into()), passed_defs.file_imports)?;
 
     // probably via macro
     // TODO dirty code collecting imports
 
     Ok(escape_var(s))
-  } else if def_ns == ns {
-    println!("[Warn] detected unresolved variable `{}` in {}/{}", s, ns, at_def);
+  } else if def_ns == passed_defs.ns {
+    println!("[Warn] detected unresolved variable `{}` in {}/{}", s, passed_defs.ns, at_def);
     Ok(escape_var(s))
   } else {
-    println!("[Warn] Unexpected case, code gen for `{}` in {}/{}", s, ns, at_def);
+    println!("[Warn] Unexpected case, code gen for `{}` in {}/{}", s, passed_defs.ns, at_def);
     Ok(format!("{}{}", var_prefix, escape_var(s)))
   }
 }
@@ -671,7 +687,7 @@ fn gen_let_code(
   xs: &Calcit,
   ns: &str,
   file_imports: &RefCell<ImportsDict>,
-  keywords: &RefCell<Vec<EdnKwd>>,
+  keywords: &RefCell<HashSet<EdnKwd>>,
   base_return_label: Option<&str>,
 ) -> Result<String, String> {
   let mut let_def_body = body.to_owned();
@@ -790,7 +806,7 @@ fn gen_if_code(
   _xs: &Calcit,
   ns: &str,
   file_imports: &RefCell<ImportsDict>,
-  keywords: &RefCell<Vec<EdnKwd>>,
+  keywords: &RefCell<HashSet<EdnKwd>>,
   base_return_label: Option<&str>,
 ) -> Result<String, String> {
   if body.len() < 2 || body.len() > 3 {
@@ -848,7 +864,7 @@ fn gen_args_code(
   ns: &str,
   local_defs: &HashSet<Arc<str>>,
   file_imports: &RefCell<ImportsDict>,
-  keywords: &RefCell<Vec<EdnKwd>>,
+  keywords: &RefCell<HashSet<EdnKwd>>,
 ) -> Result<String, String> {
   let mut result = String::from("");
   let var_prefix = if ns == "calcit.core" { "" } else { "$calcit." };
@@ -884,7 +900,7 @@ fn list_to_js_code(
   local_defs: HashSet<Arc<str>>,
   return_label: &str,
   file_imports: &RefCell<ImportsDict>,
-  keywords: &RefCell<Vec<EdnKwd>>,
+  keywords: &RefCell<HashSet<EdnKwd>>,
 ) -> Result<String, String> {
   // TODO default returnLabel="return "
   let mut result = String::from("");
@@ -931,14 +947,12 @@ fn gen_js_func(
   name: &str,
   args: &[Arc<str>],
   raw_body: &CalcitItems,
-  ns: &str,
+  passed_defs: &PassedDefs,
   exported: bool,
-  outer_defs: &HashSet<Arc<str>>,
-  file_imports: &RefCell<ImportsDict>,
-  keywords: &RefCell<Vec<EdnKwd>>,
+  keywords: &RefCell<HashSet<EdnKwd>>,
 ) -> Result<String, String> {
-  let var_prefix = if ns == "calcit.core" { "" } else { "$calcit." };
-  let mut local_defs = outer_defs.to_owned();
+  let var_prefix = if passed_defs.ns == "calcit.core" { "" } else { "$calcit." };
+  let mut local_defs = passed_defs.local_defs.to_owned();
   let mut spreading_code = String::from(""); // js list and calcit-js list are different, need to convert
   let mut args_code = String::from("");
   let mut spreading = false;
@@ -987,7 +1001,7 @@ fn gen_js_func(
   } else if has_optional {
     snippets::tmpl_args_between(args_count, args_count + optional_count)
   } else {
-    snippets::tmpl_args_exact(args_count)
+    snippets::tmpl_args_exact(name, args_count)
   };
 
   let mut body: CalcitItems = TernaryTreeList::Empty;
@@ -1009,7 +1023,14 @@ fn gen_js_func(
 
   if !body.is_empty() && uses_recur(&body[body.len() - 1]) {
     let return_var = js_gensym("return_mark");
-    let body = list_to_js_code(&body, ns, local_defs, &format!("%%{}%% =", return_var), file_imports, keywords)?;
+    let body = list_to_js_code(
+      &body,
+      passed_defs.ns,
+      local_defs,
+      &format!("%%{}%% =", return_var),
+      passed_defs.file_imports,
+      keywords,
+    )?;
     let fn_def = snippets::tmpl_tail_recursion(
       /* name = */ escape_var(name),
       /* args_code = */ args_code,
@@ -1032,13 +1053,13 @@ fn gen_js_func(
     Ok(format!("{}{}\n", export_mark, fn_def))
   } else {
     let fn_definition = format!(
-      "{}function {}({}) {{ {}{}\n{} }}",
+      "{}function {}({}) {{ {}{}\n{}}}",
       async_prefix,
       escape_var(name),
       args_code,
       check_args,
       spreading_code,
-      list_to_js_code(&body, ns, local_defs, "return ", file_imports, keywords)?
+      list_to_js_code(&body, passed_defs.ns, local_defs, "return ", passed_defs.file_imports, keywords)?
     );
     let export_mark = if exported { "export " } else { "" };
     Ok(format!("{}{}\n", export_mark, fn_definition))
@@ -1153,7 +1174,7 @@ pub fn emit_js(entry_ns: &str, emit_path: &str) -> Result<(), String> {
     // side-effects, reset tracking state
 
     let file_imports: RefCell<ImportsDict> = RefCell::new(BTreeMap::new());
-    let keywords: RefCell<Vec<EdnKwd>> = RefCell::new(Vec::new());
+    let keywords: RefCell<HashSet<EdnKwd>> = RefCell::new(HashSet::new());
 
     let mut defs_in_current: HashSet<Arc<str>> = HashSet::new();
     for k in file.keys() {
@@ -1230,7 +1251,12 @@ pub fn emit_js(entry_ns: &str, emit_path: &str) -> Result<(), String> {
           ..
         } => {
           gen_stack::push_call_stack(def_ns, name, StackKind::Codegen, f.to_owned(), &TernaryTreeList::Empty);
-          defs_code.push_str(&gen_js_func(&def, args, code, &ns, true, &def_names, &file_imports, &keywords)?);
+          let passed_defs = PassedDefs {
+            ns: &ns,
+            local_defs: &def_names,
+            file_imports: &file_imports,
+          };
+          defs_code.push_str(&gen_js_func(&def, args, code, &passed_defs, true, &keywords)?);
           gen_stack::pop_call_stack();
         }
         Calcit::Thunk(code, _) => {
@@ -1302,11 +1328,13 @@ pub fn emit_js(entry_ns: &str, emit_path: &str) -> Result<(), String> {
 
     let kwd_prefix = if &*ns == "calcit.core" { "" } else { "$calcit." };
     let mut kwd_arr = String::from("[");
-    {
-      // need to maintain a stable order to reduce redundant reloads
-      let mut kwds = keywords.borrow_mut();
-      kwds.dedup();
+    let mut kwds: Vec<EdnKwd> = vec![];
+    for k in keywords.borrow().iter() {
+      kwds.push(k.to_owned());
     }
+    // need to maintain a stable order to reduce redundant reloads
+    kwds.sort();
+
     for s in keywords.into_inner() {
       let name = escape_cirru_str(&s.to_string());
       kwd_arr.push_str(&format!("{},", name));
@@ -1318,7 +1346,15 @@ pub fn emit_js(entry_ns: &str, emit_path: &str) -> Result<(), String> {
     let js_file_path = code_emit_path.join(to_js_file_name(&ns, false)); // TODO mjs_mode
     let wrote_new = write_file_if_changed(
       &js_file_path,
-      &format!("{}\n{}{}\n\n{}\n{}", import_code, keywords_code, defs_code, vals_code, direct_code),
+      &format!(
+        "{}{}{}\n{}\n\n{}\n{}",
+        import_code,
+        keywords_code,
+        snippets::tmpl_errors_init(),
+        defs_code,
+        vals_code,
+        direct_code
+      ),
     );
     if wrote_new {
       println!("Emitted js file: {}", js_file_path.to_str().unwrap());
